@@ -3,6 +3,7 @@ import glob
 import io
 import os
 import shutil
+import sys
 import threading
 from shlex import quote
 from typing import (
@@ -69,6 +70,13 @@ def get_filesystem_class(
     return _proto2filesystem_class[proto]
 
 
+def list_filesystems() -> List[Type["FileSystem"]]:
+    """
+    Returns list of supported filesystems.
+    """
+    return list(_proto2filesystem_class.values())
+
+
 def open_file(url: str, mode: str = "r", encoding: Optional[str] = None, **kwargs: Any) -> IO:
     """
     Open a file stream.
@@ -87,6 +95,49 @@ def open_file(url: str, mode: str = "r", encoding: Optional[str] = None, **kwarg
     return get_filesystem(url=url).open(url, mode=mode, encoding=encoding, **kwargs)
 
 
+def copy_file(src_path: Optional[str], dest_path: Optional[str], recursive: bool = False) -> None:
+    """
+    Copy a file. Files can be from different filesystems.
+
+    Parameters
+    ----------
+    src_path : Optional[str]
+        Source path to copy from. If None, use stdin.
+    dest_path : Optional[str]
+        Destination path to copy to. If None, use stdout.
+    recursive : bool
+        If True, copy a directory tree of files.
+    """
+    if src_path is not None and dest_path is not None:
+        if not recursive:
+            File(src_path).copy_to(File(dest_path))
+        else:
+            Dir(src_path).copy_to(Dir(dest_path))
+        return
+
+    if recursive:
+        raise ValueError("recursive is not supported with stdin or stdout.")
+
+    infile: Optional[IO] = None
+    outfile: Optional[IO] = None
+    try:
+        if src_path is None:
+            infile = sys.stdin.buffer
+        else:
+            infile = open_file(src_path, "rb")
+        if dest_path is None:
+            outfile = sys.stdout.buffer
+        else:
+            outfile = open_file(dest_path, "wb")
+
+        shutil.copyfileobj(infile, outfile)
+    finally:
+        if infile:
+            infile.close()
+        if outfile:
+            outfile.close()
+
+
 def glob_file(pattern: str) -> List[str]:
     return get_filesystem(url=pattern).glob(pattern)
 
@@ -194,6 +245,41 @@ class FileSystem(abc.ABC):
         with open_file(src_path, "rb") as infile, open_file(dest_path, "wb") as outfile:
             shutil.copyfileobj(infile, outfile)
 
+    def shell_copy(
+        self, src_path: Optional[str], dest_path: Optional[str], recursive: bool = False
+    ) -> str:
+        """
+        Returns a shell command for performing a file copy.
+
+        Parameters
+        ----------
+        src_path : Optional[str]
+            Source path to copy from. If None, use stdin.
+        dest_path : Optional[str]
+            Destination path to copy to. If None, use stdout.
+        recursive : bool
+            If True, copy a directory tree of files.
+        """
+        if recursive:
+            if src_path and dest_path:
+                return f"redun fs cp --recursive {quote(src_path)} {quote(dest_path)}"
+            else:
+                raise ValueError("recursive is not supported with stdin or stdout.")
+
+        src_cmd = get_filesystem(url=src_path).shell_copy(src_path, None) if src_path else None
+        dest_cmd = get_filesystem(url=dest_path).shell_copy(None, dest_path) if dest_path else None
+
+        if src_cmd and dest_cmd:
+            # Combine two fs-specific shell commands through a pipe.
+            return f"({src_cmd}) | ({dest_cmd})"
+        elif src_cmd:
+            return src_cmd
+        elif dest_cmd:
+            return dest_cmd
+        else:
+            # At least one path must be defined.
+            raise ValueError("At least one path must be defined.")
+
     def glob(self, pattern: str) -> List[str]:
         """
         Returns filenames matching pattern.
@@ -308,6 +394,45 @@ class LocalFileSystem(FileSystem):
             # Perform generic copy.
             super().copy(src_path, dest_path)
 
+    def shell_copy(
+        self, src_path: Optional[str], dest_path: Optional[str], recursive: bool = False
+    ) -> str:
+        """
+        Returns a shell command for performing a file copy.
+
+        Parameters
+        ----------
+        src_path : Optional[str]
+            Source path to copy from. If None, use stdin.
+        dest_path : Optional[str]
+            Destination path to copy to. If None, use stdout.
+        recursive : bool
+            If True, copy a directory tree of files.
+        """
+        protos = {get_proto(path) for path in [src_path, dest_path] if path}
+        if "local" not in protos:
+            # Fallback to generic copy strategy.
+            return super().shell_copy(src_path, dest_path, recursive=recursive)
+
+        other_proto = next((proto for proto in protos if proto != "local"), None)
+        if other_proto:
+            # Let other proto produce the command.
+            return get_filesystem(other_proto).shell_copy(src_path, dest_path, recursive=recursive)
+
+        if src_path and dest_path:
+            if not recursive:
+                return f"cp {quote(src_path)} {quote(dest_path)}"
+            else:
+                return f"cp -r {quote(src_path)} {quote(dest_path)}"
+        elif recursive:
+            raise ValueError("recursive is not supported with stdin or stdout.")
+        elif src_path:
+            return f"cat {quote(src_path)}"
+        elif dest_path:
+            return f"cat > {quote(dest_path)}"
+        else:
+            raise ValueError("At least one path must be given.")
+
     def glob(self, pattern: str) -> List[str]:
         """
         Returns filenames matching pattern.
@@ -408,6 +533,30 @@ class FsspecFileSystem(FileSystem):
             # Perform generic copy.
             super().copy(src_path, dest_path)
 
+    def shell_copy(
+        self, src_path: Optional[str], dest_path: Optional[str], recursive: bool = False
+    ) -> str:
+        """
+        Returns a shell command for performing a file copy.
+
+        Parameters
+        ----------
+        src_path : Optional[str]
+            Source path to copy from. If None, use stdin.
+        dest_path : Optional[str]
+            Destination path to copy to. If None, use stdout.
+        recursive : bool
+            If True, copy a directory tree of files.
+        """
+        if recursive:
+            if not src_path or not dest_path:
+                raise ValueError("recursive is not supported with stdin or stdout.")
+            return f"redun fs cp --recursive {quote(src_path)} {quote(dest_path)}"
+        else:
+            src_path = src_path or "-"
+            dest_path = dest_path or "-"
+            return f"redun fs cp {quote(src_path)} {quote(dest_path)}"
+
     def glob(self, pattern: str) -> List[str]:
         """
         Returns filenames matching pattern.
@@ -468,6 +617,40 @@ class GSFileSystem(FsspecFileSystem):
 
     name = "gs"
 
+    def shell_copy(
+        self, src_path: Optional[str], dest_path: Optional[str], recursive: bool = False
+    ) -> str:
+        """
+        Returns a shell command for performing a file copy.
+
+        Parameters
+        ----------
+        src_path : Optional[str]
+            Source path to copy from. If None, use stdin.
+        dest_path : Optional[str]
+            Destination path to copy to. If None, use stdout.
+        recursive : bool
+            If True, copy a directory tree of files.
+        """
+        protos = {get_proto(path) for path in [src_path, dest_path] if path}
+        if "gs" not in protos or not (protos <= {"local", "gs"}):
+            # Fallback to generic copy strategy.
+            return super().shell_copy(src_path, dest_path, recursive=recursive)
+
+        if src_path and dest_path:
+            if not recursive:
+                return f"gsutil cp {quote(src_path)} {quote(dest_path)}"
+            else:
+                return f"gsutil cp -r {quote(src_path)} {quote(dest_path)}"
+        elif recursive:
+            raise ValueError("recursive is not supported with stdin or stdout.")
+        elif src_path:
+            return f"gsutil cp {quote(src_path)} -"
+        elif dest_path:
+            return f"gsutil cp - {quote(dest_path)}"
+        else:
+            raise ValueError("At least one path must be given.")
+
 
 @register_filesystem
 class S3FileSystem(FileSystem):
@@ -564,6 +747,40 @@ class S3FileSystem(FileSystem):
             # Perform generic copy.
             super().copy(src_path, dest_path)
 
+    def shell_copy(
+        self, src_path: Optional[str], dest_path: Optional[str], recursive: bool = False
+    ) -> str:
+        """
+        Returns a shell command for performing a file copy.
+
+        Parameters
+        ----------
+        src_path : Optional[str]
+            Source path to copy from. If None, use stdin.
+        dest_path : Optional[str]
+            Destination path to copy to. If None, use stdout.
+        recursive : bool
+            If True, copy a directory tree of files.
+        """
+        protos = {get_proto(path) for path in [src_path, dest_path] if path}
+        if "s3" not in protos or not (protos <= {"local", "s3"}):
+            # Fallback to generic copy strategy.
+            return super().shell_copy(src_path, dest_path, recursive=recursive)
+
+        if src_path and dest_path:
+            if not recursive:
+                return f"aws s3 cp --no-progress {quote(src_path)} {quote(dest_path)}"
+            else:
+                return f"aws s3 cp --no-progress --recursive {quote(src_path)} {quote(dest_path)}"
+        elif recursive:
+            raise ValueError("recursive is not supported with stdin or stdout.")
+        elif src_path:
+            return f"aws s3 cp --no-progress {quote(src_path)} -"
+        elif dest_path:
+            return f"aws s3 cp --no-progress - {quote(dest_path)}"
+        else:
+            raise ValueError("At least one path must be given.")
+
     def glob(self, pattern: str) -> List[str]:
         return ["s3://" + key for key in self.s3.glob(pattern)]
 
@@ -688,6 +905,17 @@ class File(Value):
         dest_file.update_hash()
         return dest_file
 
+    def shell_copy_to(self, dest_path: Optional[str]) -> str:
+        """
+        Returns a shell command for copying the file to a destination path.
+
+        Parameters
+        ----------
+        dest_path : Optional[str]
+            Destination path to copy to. If None, use stdout.
+        """
+        return self.filesystem.shell_copy(self.path, dest_path)
+
     def isfile(self) -> bool:
         return self.filesystem.isfile(self.path)
 
@@ -835,6 +1063,17 @@ class Dir(FileSet):
             src_file.copy_to(dest_file, skip_if_exists=skip_if_exists)
         return dest_dir
 
+    def shell_copy_to(self, dest_path: str) -> str:
+        """
+        Returns a shell command for copying the directory to a destination path.
+
+        Parameters
+        ----------
+        dest_path : str
+            Destination path to copy to.
+        """
+        return self.filesystem.shell_copy(self.path, dest_path, recursive=True)
+
     def stage(self, local: Optional[str] = None) -> "StagingDir":
         if not local:
             local = os.path.basename(self.path)
@@ -1191,6 +1430,77 @@ class ShardedS3Dataset(Value):
         context.purge_s3_path(self.path, options=options)
         self.update_hash()
 
+    @classmethod
+    def from_files(
+        cls, files: List[File], format: Optional[str] = None, allow_additional_files: bool = False
+    ) -> "ShardedS3Dataset":
+        """
+        Helper function to create a ShardedS3Dataset from a list of redun Files.
+        This can be helpful for provenance tracking from methods that return multiple
+        files and the "root directory" is wanted for later use.
+
+        If a dataset cannot be created that contains all the files, such as being in
+        different S3 prefixes, a ValueError is raised. If files do not all have the
+        same format, a ValueError will be raised.
+
+        Parameters
+        ----------
+        files : List[File]
+            Files that should be included in the dataset.
+
+        format : Optional[str]
+            If specified, files are considered to be in this format.
+
+        allow_additional_files : bool
+            If True, allows the resulting ShardedS3Dataset to contain files not in
+            `files`. If False, a ValueError will be raised if a ShardedS3Dataset cannot
+            be constructed.
+
+        Raises
+        ------
+        ValueError
+            If a dataset could not be created from the given files.
+        """
+        # Check all files are actually on S3.
+        if not all(f.filesystem.name == "s3" for f in files):
+            raise ValueError("All files must be on S3 to construct a ShardedS3Dataset")
+
+        # All files should exist.
+        for f in files:
+            if not f.exists():
+                raise FileNotFoundError(
+                    f"File {f.path} must exist to construct a dataset with it."
+                )
+
+        # os.path does not like the "s3://" prefix so we remove it and then add it back.
+        if len(files) == 1:
+            root_path = f"s3:/{os.path.split(files[0].path.replace('s3://', '/'))[0]}"
+        elif len(files) > 1:
+            root_path = f"s3:/{os.path.commonpath([f.path.replace('s3://', '/') for f in files])}"
+        else:
+            raise ValueError("Cannot construct a dataset from an empty list of files.")
+
+        if not format:
+            formats = set(os.path.splitext(f.path)[1].replace(".", "") for f in files)
+            if not formats or len(formats) > 1:
+                raise ValueError(
+                    f"Multiple or zero file extensions found: {formats}."
+                    " Could not detect format for ShardedS3Dataset."
+                )
+            format = formats.pop()
+
+        result = ShardedS3Dataset(root_path, format=format, recurse=True)
+
+        # Other files should not be included in the dataset unless specifically requested.
+        if not allow_additional_files:
+            extras = set(result.filenames) - set(f.path for f in files)
+            if extras:
+                raise ValueError(
+                    f"Additional files {extras} would be added to dataset but "
+                    "`allow_additional_files=False`"
+                )
+        return result
+
     @classmethod
     def from_data(
         cls,
@@ -1201,7 +1511,7 @@ class ShardedS3Dataset(Value):
         catalog_database: str = "default",
         catalog_table: Optional[str] = None,
         format_options: Dict[str, Any] = {},
-    ):
+    ) -> "ShardedS3Dataset":
         """
         Helper function to create a ShardedS3Dataset from an existing DataFrame-like object.
 
@@ -1315,30 +1625,24 @@ class StagingFile(Staging[File]):
         return self.local.copy_to(self.remote)
 
     def render_unstage(self) -> str:
+        """
+        Returns a shell command for unstaging a file.
+        """
         if self.local.path == self.remote.path:
             # No staging is needed.
             return ""
 
-        # TODO: Generalize.
-        if self.remote.filesystem.name == "s3" or self.local.filesystem.name == "s3":
-            return "aws s3 cp --no-progress {} {}".format(
-                quote(self.local.path), quote(self.remote.path)
-            )
-        else:
-            return "cp {} {}".format(quote(self.local.path), quote(self.remote.path))
+        return self.local.shell_copy_to(self.remote.path)
 
     def render_stage(self) -> str:
+        """
+        Returns a shell command for staging a file.
+        """
         if self.local.path == self.remote.path:
             # No staging is needed.
             return ""
 
-        # TODO: Generalize.
-        if self.remote.filesystem.name == "s3" or self.local.filesystem.name == "s3":
-            return "aws s3 cp --no-progress {} {}".format(
-                quote(self.remote.path), quote(self.local.path)
-            )
-        else:
-            return "cp {} {}".format(quote(self.remote.path), quote(self.local.path))
+        return self.remote.shell_copy_to(self.local.path)
 
 
 class StagingDir(Staging[Dir]):
@@ -1383,27 +1687,21 @@ class StagingDir(Staging[Dir]):
         return self.local.copy_to(self.remote)
 
     def render_unstage(self) -> str:
+        """
+        Returns a shell command for unstaging a directory.
+        """
         if self.local.path == self.remote.path:
             # No staging is needed.
             return ""
 
-        # TODO: Generalize.
-        if self.remote.filesystem.name == "s3" or self.local.filesystem.name == "s3":
-            return "aws s3 cp --no-progress --recursive {} {}".format(
-                quote(self.local.path), quote(self.remote.path)
-            )
-        else:
-            return "cp -r {} {}".format(quote(self.local.path), quote(self.remote.path))
+        return self.local.shell_copy_to(self.remote.path)
 
     def render_stage(self) -> str:
+        """
+        Returns a shell command for staging a directory.
+        """
         if self.local.path == self.remote.path:
             # No staging is needed.
             return ""
 
-        # TODO: Generalize.
-        if self.remote.filesystem.name == "s3" or self.local.filesystem.name == "s3":
-            return "aws s3 cp --no-progress --recursive {} {}".format(
-                quote(self.remote.path), quote(self.local.path)
-            )
-        else:
-            return "cp -r {} {}".format(quote(self.remote.path), quote(self.local.path))
+        return self.remote.shell_copy_to(self.local.path)
