@@ -288,7 +288,13 @@ def test_get_or_create_job_definition(get_job_definition_mock, get_aws_client_mo
                         "mountPoints": [],
                         "volumes": [],
                         "resourceRequirements": [],
-                        "ulimits": [],
+                        "ulimits": [
+                            {
+                                "name": "nofile",
+                                "softLimit": 65535,
+                                "hardLimit": 65535,
+                            }
+                        ],
                         "privileged": False,
                     },
                     "targetNodes": "0",
@@ -304,7 +310,13 @@ def test_get_or_create_job_definition(get_job_definition_mock, get_aws_client_mo
                         "mountPoints": [],
                         "volumes": [],
                         "resourceRequirements": [],
-                        "ulimits": [],
+                        "ulimits": [
+                            {
+                                "name": "nofile",
+                                "softLimit": 65535,
+                                "hardLimit": 65535,
+                            }
+                        ],
                         "privileged": False,
                     },
                     "targetNodes": "1:",
@@ -1371,6 +1383,7 @@ def test_executor_docker(
             "shared_memory": None,
             "vcpus": 1,
             "volumes": [(scratch_dir, scratch_dir)],
+            "interactive": False,
         }
 
         run_docker_mock.reset_mock()
@@ -1394,6 +1407,7 @@ def test_executor_docker(
             "shared_memory": None,
             "vcpus": 1,
             "volumes": [(scratch_dir, scratch_dir)],
+            "interactive": False,
         }
 
         # Simulate output file created by job.
@@ -1468,6 +1482,7 @@ def test_executor_job_debug(
             "shared_memory": None,
             "vcpus": 1,
             "volumes": [(scratch_dir, scratch_dir)],
+            "interactive": False,
         }
 
         # Simulate output file created by job.
