@@ -1,13 +1,16 @@
+import dataclasses
 import os
 from traceback import FrameSummary
-from typing import Any, Dict, List, Sequence, Tuple
+from typing import Any, Dict, List, Optional, Sequence, Tuple
 from unittest.mock import Mock, patch
 
 import pytest
 from sqlalchemy.orm import Session
 
 from redun import Scheduler, task
-from redun.backends.db import Execution, RedunBackendDb
+from redun.backends.db import Execution
+from redun.backends.db import Job as JobDb
+from redun.backends.db import RedunBackendDb
 from redun.config import Config
 from redun.expression import SchedulerExpression
 from redun.promise import Promise
@@ -28,6 +31,12 @@ from redun.utils import map_nested_value
 from redun.value import Value, get_type_registry
 
 
+@dataclasses.dataclass
+class Node:
+    value: str
+    next_: Optional["Node"]
+
+
 def test_simple(scheduler: Scheduler) -> None:
     """
     A single task should execute.
@@ -407,6 +416,21 @@ def test_nested_results(scheduler: Scheduler) -> None:
     assert result == [10]
 
 
+def test_nested_results2(scheduler: Scheduler) -> None:
+    @task()
+    def to_str(x: int) -> str:
+        return str(x)
+
+    def create_linked_list(y: List[int]) -> Optional[Node]:
+        if len(y) > 1:
+            return Node(value=to_str(y[0]), next_=create_linked_list(y[1:]))
+        else:
+            return Node(to_str(y[0]), next_=None) if y else None
+
+    assert scheduler.run(create_linked_list([1])) == Node("1", None)
+    assert scheduler.run(create_linked_list([1, 2])) == Node("1", Node("2", None))
+
+
 def test_higher_order(scheduler: Scheduler) -> None:
     """
     Using task as an argument to another task.
@@ -703,6 +727,7 @@ def test_log_job_status(scheduler: Scheduler) -> None:
         "| ALL           0       0       0       0       3       3",
         "| main          0       0       0       0       1       1",
         "| task1         0       0       0       0       2       2",
+        "",
     ]
 
 
@@ -1531,3 +1556,190 @@ def test_job_clear(scheduler: Scheduler) -> None:
 
     with pytest.raises(ValueError, match="raised by child"):
         scheduler.run(main(10))
+
+
+def test_common_expression() -> None:
+    """
+    Scheduler should use Common Expression Elimination.
+    """
+
+    @task
+    def id(x: int) -> int:
+        return x
+
+    @task
+    def add(a: int, b: int) -> int:
+        return a + b
+
+    @task
+    def boom(x: int) -> int:
+        raise ValueError("BOOM!")
+
+    # There should be only one job for add.
+    scheduler = Scheduler()
+    assert isinstance(scheduler.backend, RedunBackendDb)
+    assert scheduler.backend.session
+    assert scheduler.run([add(1, 2), add(1, 2)]) == [3, 3]
+    assert scheduler.backend.session.query(JobDb).filter(JobDb.task_hash == add.hash).count() == 1
+
+    # We should do CSE after argument expressions are evaluated, such as `id(2)`.
+    scheduler = Scheduler()
+    assert isinstance(scheduler.backend, RedunBackendDb)
+    assert scheduler.backend.session
+
+    # Cache id(2) so that we can force `add(1, id(2))` and `add(1, 2)` at the same time
+    # and excercise the CSE logic on two different expressions.
+    assert scheduler.run(id(2)) == 2
+    assert scheduler.run([add(1, id(2)), add(1, 2)]) == [3, 3]
+    assert scheduler.backend.session.query(JobDb).filter(JobDb.task_hash == add.hash).count() == 1
+
+    # Expressions that have been collapsed by CSE should also propagate errors.
+    scheduler = Scheduler()
+    assert isinstance(scheduler.backend, RedunBackendDb)
+    assert scheduler.backend.session
+    with pytest.raises(ValueError, match="BOOM"):
+        scheduler.run([boom(1), boom(1)])
+    assert scheduler.backend.session.query(JobDb).filter(JobDb.task_hash == boom.hash).count() == 1
+
+
+def test_cycle_expression() -> None:
+    """
+    Scheduler should not double evaluate cycles in the Expression Graph.
+    """
+    calls = []
+
+    @scheduler_task(namespace="redun.tests.test_scheduler")
+    def echo(scheduler, parent_job, expr, x) -> Promise:
+        calls.append("id")
+        promise: Promise[int] = Promise()
+        promise.do_resolve(x)
+        return promise
+
+    # We should call echo only once.
+    scheduler = Scheduler()
+    x = echo(1)
+    y = x + x
+    assert scheduler.run(y) == 2
+    assert calls == ["id"]
+
+
+def test_cycle_expression_cache(scheduler: Scheduler, session: Session) -> None:
+    """
+    Detecting expression cycles should work with cached expressions.
+    """
+
+    @task
+    def task1():
+        return 10
+
+    @task
+    def add(x, y):
+        return x + y
+
+    @task
+    def main():
+        x = task1()
+
+        # Use x twice so that we have a dimond in the Expression graph.
+        return {
+            "a": add(x, 1),
+            "b": add(x, 2),
+        }
+
+    original_collapse = Job.collapse
+
+    def collapse(self, other_job):
+        nonlocal calls
+        calls += 1
+        return original_collapse(self, other_job)
+
+    calls = 0
+    with patch.object(Job, "collapse", collapse):
+        assert scheduler.run(main()) == {"a": 11, "b": 12}
+
+    # We should not create extra Jobs needing collapsing.
+    assert calls == 0
+
+    # Update a task in the workflow.
+    @task(version="1")  # type: ignore[no-redef]
+    def add(x, y):
+        return x + y
+
+    # Even when using an Expression from the cache, we should avoid needing to use Job collapsing.
+    calls = 0
+    with patch.object(Job, "collapse", collapse):
+        assert scheduler.run(main()) == {"a": 11, "b": 12}
+
+    assert calls == 0
+
+
+def test_extend_run(scheduler: Scheduler, session: Session) -> None:
+    """
+    Scheduler should support extending an Execution.
+    """
+
+    @task
+    def main(x):
+        return x + 1
+
+    @task
+    def fail():
+        raise ValueError("BOOM")
+
+    assert scheduler.run(main(10)) == 11
+    job = session.query(JobDb).one()
+
+    result = scheduler.extend_run(main(11), parent_job_id=job.id)
+    assert result["result"] == 12
+
+    # Ensure other execution info is present.
+    assert result["job_id"]
+    assert result["call_hash"]
+
+    # The new job should be a child of the parent job.
+    child_job = session.query(JobDb).get(result["job_id"])
+    assert child_job.parent_job.id == job.id
+    assert child_job.id in {j.id for j in job.child_jobs}
+
+    # extend_run should also return errors.
+    result = scheduler.extend_run(fail(), parent_job_id=job.id)
+    assert isinstance(result["error"], ValueError)
+    assert str(result["error"]) == "BOOM"
+    assert result["job_id"]
+    assert result["call_hash"]
+
+    # Force session to refetch new data.
+    job_id = job.id
+    session.expunge_all()
+
+    # The new job should be a child of the parent job.
+    child_job = session.query(JobDb).get(result["job_id"])
+    job = session.query(JobDb).get(job_id)
+    assert child_job.parent_job.id == job.id
+    assert child_job.id in {j.id for j in job.child_jobs}
+
+
+def test_extend_run_dryrun(scheduler: Scheduler, session: Session) -> None:
+    """
+    Scheduler should support extending an Execution with dryrun.
+    """
+
+    @task
+    def main(x):
+        return x + 1
+
+    @task
+    def fail():
+        raise ValueError("BOOM")
+
+    assert scheduler.run(main(10)) == 11
+    job = session.query(JobDb).one()
+
+    # We should have an incomplete run due to dryrun=True.
+    result = scheduler.extend_run(main(11), parent_job_id=job.id, dryrun=True)
+    assert result["dryrun"]
+
+    # If we finish the run and then do dryrun, we should get a complete run due to caching.
+    result = scheduler.extend_run(main(11), parent_job_id=job.id)
+    result = scheduler.extend_run(main(11), parent_job_id=job.id, dryrun=True)
+    assert result["result"] == 12
